/*
 * Copyright (c) 1995, 1996
 *	Sandro Sigala, Brescia, Italy.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Sandro Sigala.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: hash.c,v 1.5 1996/06/28 10:28:14 sandro Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "hash.h"

/*
 * Hashing function.
 */
static unsigned long
hash_table_hash(const void *data, unsigned long size, unsigned long table_size)
{
	const unsigned char *p = (const unsigned char *)data;
	unsigned long hash = 0, g;

	while (*p) {
		hash = (hash << 4) + *p++;
		if (g = hash & 0xF0000000)
			hash ^= g >> 24;
		hash &= ~g;
	}

	return (hash % table_size);
}

static struct hash_table_bucket *
build_bucket(const void *key, unsigned long size)
{
	struct hash_table_bucket *ptr;

	if ((ptr = (struct hash_table_bucket *)malloc(sizeof(struct hash_table_bucket))) == NULL)
		err(1, NULL);

	if ((ptr->key = malloc(size)) == NULL)
		err(1, NULL);

	memcpy(ptr->key, key, size);

	ptr->ksize = size;
	ptr->dsize = 0;
	ptr->data = NULL;
	ptr->next = NULL;
	ptr->prev = NULL;

	return ptr;
}

static void
free_bucket(struct hash_table_bucket *ptr)
{
	if (ptr->key != NULL)
		free(ptr->key);
	if (ptr->data != NULL)
		free(ptr->data);

	free(ptr);
}

struct hash_table *
hash_table_build(unsigned long size)
{
	struct hash_table *ptr;
	int i;

	if ((ptr = (struct hash_table *)malloc(sizeof(struct hash_table))) == NULL)
		err(1, NULL);

	ptr->table_size = size;

	if ((ptr->table = (struct hash_table_bucket **)malloc(ptr->table_size * sizeof(struct hash_table_bucket *))) == NULL)
		err(1, NULL);

	ptr->hashing_function = hash_table_hash;

	for (i = 0; i < ptr->table_size; i++)
		ptr->table[i] = NULL;

	return ptr;
}

struct hash_table *
hash_table_build_default(void)
{
	return hash_table_build(HASH_TABLE_DEFAULT_SIZE);
}

void
hash_table_set_hashing_function(struct hash_table *table, hashing_function_t function)
{
	table->hashing_function = function;
}

int
hash_table_free(struct hash_table *table)
{
	struct hash_table_bucket *bucket, *next;
	int i;

	for (i = 0; i < table->table_size; i++) {
		bucket = table->table[i];
		while (bucket != NULL) {
			next = bucket->next;
			free_bucket(bucket);
			bucket = next;
		}
	}

	free(table);

	return 0;
}

int
hash_table_store_key(struct hash_table *table, const void *key, unsigned long size)
{
	struct hash_table_bucket *bucket;
	unsigned long hash;

	if (hash_table_exists(table, key, size))
		return -1;

	hash = table->hashing_function(key, size, table->table_size);

	if (table->table[hash] == NULL)
		table->table[hash] = build_bucket(key, size);
	else {
		bucket = table->table[hash];

		while (bucket->next != NULL)
			bucket = bucket->next;

		bucket->next = build_bucket(key, size);
		bucket->next->prev = bucket;
	}

	return 0;
}

int
hash_table_exists(struct hash_table *table, const void *key, unsigned long size)
{
	struct hash_table_bucket *bucket;
	unsigned long hash;

	hash = table->hashing_function(key, size, table->table_size);

	for (bucket = table->table[hash]; bucket != NULL; bucket = bucket->next)
		if (!memcmp(bucket->key, key, size))
			return 1;

	return 0;
}

int
hash_table_delete(struct hash_table *table, const void *key, unsigned long size)
{
	struct hash_table_bucket *bucket;
	unsigned long hash;

	hash = table->hashing_function(key, size, table->table_size);

	for (bucket = table->table[hash]; bucket != NULL; bucket = bucket->next)
		if (!memcmp(bucket->key, key, size)) {
			if (bucket == table->table[hash])
				table->table[hash] = bucket->next;
			else {
				if (bucket->prev != NULL)
					bucket->prev->next = bucket->next;
				if (bucket->next != NULL)
					bucket->next->prev = bucket->prev;
			}
			free_bucket(bucket);
			return 0;
		}

	return -1;
}

int
hash_table_store_dataptr(struct hash_table *table, const void *key, unsigned long ksize, const void *data, unsigned long dsize)
{
	struct hash_table_bucket *bucket;
	unsigned long hash;

	hash = table->hashing_function(key, ksize, table->table_size);

	for (bucket = table->table[hash]; bucket != NULL; bucket = bucket->next)
		if (!memcmp(bucket->key, key, ksize)) {
			bucket->data = (void *) data;
			bucket->dsize = dsize;
			return 0;
		}

	return -1;
}

int
hash_table_store_data(struct hash_table *table, const void *key, unsigned long ksize, const void *data, unsigned long dsize)
{
	void *p = (void *)data;

	if (data != NULL && dsize > 0) {
		if ((p = malloc(dsize)) == NULL)
			err(1, NULL);
		memcpy(p, data, dsize);
	}

	return (hash_table_store_dataptr(table, key, ksize, p, dsize));
}

int
hash_table_store(struct hash_table *table, const void *key, unsigned long ksize, const void *data, unsigned long dsize)
{
	if (hash_table_exists(table, key, ksize))
		hash_table_delete(table, key, ksize);

	hash_table_store_key(table, key, ksize);

	return (hash_table_store_dataptr(table, key, ksize, data, dsize));
}

void *
hash_table_fetch(struct hash_table *table, const void *key, unsigned long size)
{
	struct hash_table_bucket *bucket;
	unsigned long hash;

	hash = table->hashing_function(key, size, table->table_size);

	for (bucket = table->table[hash]; bucket != NULL; bucket = bucket->next)
		if (!memcmp(bucket->key, key, size))
			return bucket->data;

	return NULL;
}
