/*
 * Copyright (c) 1995, 1996
 *	Sandro Sigala, Brescia, Italy.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Sandro Sigala.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: chilight.c,v 1.12 1996/06/28 14:21:57 sandro Exp $
 */

static char *rcsid = "$Id: chilight.c,v 1.12 1996/06/28 14:21:57 sandro Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <getopt.h>

#include "version.h"
#include "tokens.h"


int opt_directives;

extern char *token_buffer;
extern char *yytext;
extern FILE *yyin;

static FILE *output_file;

enum { UNBINDED = LEX_LAST_TOKEN + 1 };

#include "colors.h"

#define TABLE_SIZE 512

static struct { int fg, bg; char *escape; } color_table[TABLE_SIZE];

static int last_token = 0;

static void
color_table_set_color(int token, int fg, int bg)
{
	char buf[24];

	if (color_table[token].escape != NULL)
		free(color_table[token].escape);

	color_table[token].fg = fg;
	color_table[token].bg = bg;

	sprintf(buf, "\033[%dm", fg);
	if (bg != COLOR_UNCHANGED)
		sprintf(&buf[strlen(buf) - 1], ";%dm", bg);

	if ((color_table[token].escape = (char *)malloc(strlen(buf) + 1)) == NULL)
		err(1, NULL);
	strcpy(color_table[token].escape, buf);
}

static void
color_table_free(void)
{
	int i;
	for (i = 0; i < TABLE_SIZE; i++)
		if (color_table[i].escape != NULL)
			free(color_table[i].escape);
}

static void
color_table_init(void)
{
	int i;
	for (i = 0; i < TABLE_SIZE; i++) { 
		color_table[i].fg = 0;
		color_table[i].bg = 0;
		color_table[i].escape = NULL;
	}
}

static void
color_table_setup(void)
{
	int i = 0;
	while (def_color_table[i].token != -1) {
		color_table_set_color(def_color_table[i].token,
				      def_color_table[i].fg,
				      def_color_table[i].bg);
		i++;
	}
}

static void
outstr(char *s)
{
	while (*s)
		fputc(*s++, output_file);
}

static void
parse(void)
{
	int tk;

	while ((tk = yylex()) != 0) {
		if (color_table[tk].escape != NULL) {
			if (color_table[tk].fg != color_table[last_token].fg ||
			    color_table[tk].bg != color_table[last_token].bg)
				outstr(color_table[tk].escape);
		} else
			outstr(color_table[UNBINDED].escape);

		switch (tk) {
		case COMMENT: case DIRECTIVE:
		case STRING:
			outstr(token_buffer);
			break;
		case CHARACTER:
		case KEYWORD: case IDENTIFIER:
		case CONSTANT:
		case OPERATOR:
		case ELLIPSIS:
			outstr(yytext); break;
		default:
			putc(tk, output_file);
		}
		last_token = tk;
	}
}

static void
process_file(char *filename)
{
	if (filename != NULL && strcmp(filename, "-") != 0) {
		if ((yyin = fopen(filename, "r")) == NULL)
			err(1, "%s", filename);
	} else
		yyin = stdin;

	init_lex();
	parse();
	done_lex();

	if (yyin != stdin)
		fclose(yyin);
}

static void
usage(void)
{
	fprintf(stderr, "usage: chilight [-dV] [-o file] [file...]\n");
	exit(1);
}

int
main (int argc, char **argv)
{
	int i, c;

	color_table_init();
	color_table_setup();

	output_file = stdout;

	while ((c = getopt(argc, argv, "do:V")) != EOF)
		switch (c) {
		case 'd':
			opt_directives = 1;
			break;
		case 'o':
			if (output_file != stdout)
				fclose(output_file);
			if ((output_file = fopen(optarg, "w")) == NULL)
				err(1, "%s", optarg);
			break;
		case 'V':
			fprintf(stderr, "%s - %s\n", version_string, rcsid);
			exit(0);
		case '?':
		default:
			usage();
			/* NOTREACHED */
		}
	argc -= optind;
	argv += optind;

	if (argc < 1)
		process_file(NULL);
	else
		for (i = 0; i < argc; i++)
			process_file(argv[i]);

	if (output_file != stdout)
		fclose(output_file);
	color_table_free();

	return 0;
}
