/*
 * Copyright (c) 1995, 1996
 *	Sandro Sigala, Brescia, Italy.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Sandro Sigala.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: lexer.l,v 1.6 1996/06/28 10:28:18 sandro Exp $
 */

Digit		[0-9]
Literal		[a-zA-Z_]
Hex		[a-fA-F0-9]
Exp		[Ee][+-]?{Digit}+
FS		(f|F|l|L)
IS		(u|U|l|L)+

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "tokens.h"

extern int opt_directives;

static int yywrap(void);
static void string(void);
static void comment(void);
static int directive(void);
%}

%%
"/*"					{ comment(); return(COMMENT); }
"#"					{ return(directive()); }

"auto"					{ return(KEYWORD); }
"break"					{ return(KEYWORD); }
"case"					{ return(KEYWORD); }
"char"					{ return(KEYWORD); }
"const"					{ return(KEYWORD); }
"continue"				{ return(KEYWORD); }
"default"				{ return(KEYWORD); }
"do"					{ return(KEYWORD); }
"double"				{ return(KEYWORD); }
"else"					{ return(KEYWORD); }
"enum"					{ return(KEYWORD); }
"extern"				{ return(KEYWORD); }
"float"					{ return(KEYWORD); }
"for"					{ return(KEYWORD); }
"goto"					{ return(KEYWORD); }
"if"					{ return(KEYWORD); }
"int"					{ return(KEYWORD); }
"long"					{ return(KEYWORD); }
"register"				{ return(KEYWORD); }
"return"				{ return(KEYWORD); }
"short"					{ return(KEYWORD); }
"signed"				{ return(KEYWORD); }
"sizeof"				{ return(KEYWORD); }
"static"				{ return(KEYWORD); }
"struct"				{ return(KEYWORD); }
"switch"				{ return(KEYWORD); }
"typedef"				{ return(KEYWORD); }
"union"					{ return(KEYWORD); }
"unsigned"				{ return(KEYWORD); }
"void"					{ return(KEYWORD); }
"volatile"				{ return(KEYWORD); }
"while"					{ return(KEYWORD); }

{Literal}({Literal}|{Digit})*		{ return(IDENTIFIER); }

0[xX]{Hex}+{IS}?			{ return(CONSTANT); }
0{Digit}+{IS}?				{ return(CONSTANT); }
{Digit}+{IS}				{ return(CONSTANT); }
{Digit}+				{ return(CONSTANT); }
'(\\.|[^\\'])+'				{ return(CHARACTER); }

{Digit}+{Exp}{FS}?			{ return(CONSTANT); }
{Digit}*"."{Digit}+({Exp})?{FS}?	{ return(CONSTANT); }
{Digit}+"."{Digit}*({Exp})?{FS}?	{ return(CONSTANT); }

"\""					{ string(); return(STRING); }

">>="					{ return(OPERATOR); }
"<<="					{ return(OPERATOR); }
"+="					{ return(OPERATOR); }
"-="					{ return(OPERATOR); }
"*="					{ return(OPERATOR); }
"/="					{ return(OPERATOR); }
"%="					{ return(OPERATOR); }
"&="					{ return(OPERATOR); }
"^="					{ return(OPERATOR); }
"|="					{ return(OPERATOR); }
">>"					{ return(OPERATOR); }
"<<"					{ return(OPERATOR); }
"++"					{ return(OPERATOR); }
"--"					{ return(OPERATOR); }
"->"					{ return(OPERATOR); }
"&&"					{ return(OPERATOR); }
"||"					{ return(OPERATOR); }
"<="					{ return(OPERATOR); }
">="					{ return(OPERATOR); }
"=="					{ return(OPERATOR); }
"!="					{ return(OPERATOR); }
"..."					{ return(ELLIPSIS); }
";"					{ return(';'); }
"{"					{ return('{'); }
"}"					{ return('}'); }
","					{ return(','); }
":"					{ return(':'); }
"="					{ return('='); }
"("					{ return('('); }
")"					{ return(')'); }
"["					{ return('['); }
"]"					{ return(']'); }
"."					{ return('.'); }
"&"					{ return('&'); }
"!"					{ return('!'); }
"~"					{ return('~'); }
"-"					{ return('-'); }
"+"					{ return('+'); }
"*"					{ return('*'); }
"/"					{ return('/'); }
"%"					{ return('%'); }
"<"					{ return('<'); }
">"					{ return('>'); }
"^"					{ return('^'); }
"|"					{ return('|'); }
"?"					{ return('?'); }

[ \t\v\n\f]				{ return(yytext[0]); }
.					{ return(yytext[0]); }

%%

char *token_buffer;

static int maxtoken;

void
init_lex(void)
{
	maxtoken = 40;
	if ((token_buffer = (char *)malloc(maxtoken + 1)) == NULL)
		err(1, NULL);
}

void
done_lex(void)
{
	free(token_buffer);
}

static char *
extend_token_buffer(char *p)
{
	int offset = p - token_buffer;

	maxtoken = maxtoken * 2 + 10;

	if ((token_buffer = (char *)realloc(token_buffer, maxtoken + 2)) == NULL)
		err(1, NULL);

	return token_buffer + offset;
}

static int
yywrap(void)
{
	return 1;
}

static void
string(void)
{
	char *p;
	int c1;

	p = token_buffer;
	*p++ = '"';
	while ((c1 = input()) && c1 != '"' && c1 != EOF) {
		if (p >= token_buffer + maxtoken)
			p = extend_token_buffer(p);
		*p++ = c1;
		if (c1 == '\\')
			*p++ = input();
	}
	if (c1 == EOF)
		errx(1, "unexpected end of file in string");
	*p++ = '"';
	*p = '\0';
}

static void
comment(void)
{
	char *p;
	int c1, c2;

	p = token_buffer;
	*p++ = '/';
	*p++ = '*';
	while (c1 = input()) {
		if (p >= token_buffer + maxtoken)
			p = extend_token_buffer(p);
		*p++ = c1;
		if (c1 == '*')
			if ((c2 = input()) == '/') {
				*p++ = c2;
				*p = '\0';
				return;
			} else
				*p++ = c2;
	}
	*p = '\0';
}

static int
directive(void)
{
	char *p;
	int c1, c2;

	if (!opt_directives) {
		p = token_buffer;
		*p++ = '#';

		while ((c1 = input()) && c1 != '\n') {
			if (p >= token_buffer + maxtoken)
					p = extend_token_buffer(p);
			*p++ = c1;
			if (c1 == '\\') {
				*p++ = input();
			}
		}
		*p++ = c1;
		*p = '\0';
		return DIRECTIVE;
	} else
		return '#';
}
