/*
 * Copyright (c) 1995, 1996
 *	Sandro Sigala, Brescia, Italy.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Sandro Sigala.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cinfo.c,v 1.20 1996/06/29 16:20:46 sandro Exp $
 */

static char *rcsid = "$Id: cinfo.c,v 1.20 1996/06/29 16:20:46 sandro Exp $";

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <dirent.h>
#include <getopt.h>

#include "version.h"
#include "pattern.h"
#include "cinfodb.h"


#ifndef CINFODBDIR
#define CINFODBDIR "/usr/libexec/cinfo"
#endif

/*
 * The variables where are stored the options flags.
 */
static int opt_case;
static int opt_debug;
static int opt_list;
static int opt_pattern;
static int opt_cinfodb_directory;
static char *opt_cinfodb_directory_arg;

static FILE *output_file;

static struct object_s *
make_object_tree(char *dbdir)
{
	DIR *dir;
	struct dirent *de;
	struct object_s *op, *head = NULL, *last;
	char *p;
	int i = 0;

	if (opt_debug)
		printf("scanning directory %s\n", dbdir);

	if ((dir = opendir(dbdir)) == NULL)
		err(1, "%s", dbdir);
	while ((de = readdir(dir)) != NULL) {
		if (opt_debug)
			printf("checking %s -> ", de->d_name);
		if (strlen(de->d_name) > 8 && !strcmp(de->d_name + strlen(de->d_name) - 8, ".cinfodb")) {
			if (opt_debug)
				printf("ok\n");
			i++;
			if ((p = (char *)malloc(strlen(dbdir) + strlen(de->d_name) + 2)) == NULL)
				err(1, NULL);
			strcpy(p, dbdir);
			strcat(p, "/");
			strcat(p, de->d_name);
			if (opt_debug)
				printf("loading %s...", p);
			op = read_cinfodb(p);
			free(p);
			if (opt_debug)
				printf(" done\n");

			if (head == NULL)
				head = op;
			else
				last->next = op;
			last = op;
		} else if (opt_debug)
			printf("no\n");
	}
	closedir(dir);

	if (i == 0)
		errx(1, "%s: no cinfo database file found", dbdir);

	return head;
}

struct	string_s {
	char *s;
	struct string_s *prev;
	struct string_s *next;
};

static struct string_s *string_stack_head;
static struct string_s *string_stack_tail;

static void
string_push(char *s)
{
	struct string_s *sp;

	if ((sp = (struct string_s *)malloc(sizeof(struct string_s))) == NULL)
		err(1, NULL);
	if ((sp->s = (char *)malloc(strlen(s) + 1)) == NULL)
		err(1, NULL);
	strcpy(sp->s, s);

	sp->next = NULL;
	sp->prev = NULL;
	
	if (string_stack_head == NULL)
		string_stack_head = sp;
	else {
		string_stack_tail->next = sp;
		sp->prev = string_stack_tail;
	}
	string_stack_tail = sp;
}

static void
string_pop(void)
{
	struct string_s *sp;
	if (string_stack_head == string_stack_tail) {
		free(string_stack_head);
		string_stack_head = string_stack_tail = NULL;
	} else {
		sp = string_stack_tail->prev;
		free(string_stack_tail);
		sp->next = NULL;
		string_stack_tail = sp;
	}
}

static void
output_string_stack_tree(void)
{
	struct string_s *sp = string_stack_head;
	fprintf(output_file, "[");
	for (; sp != NULL; sp = sp->next) {
		if (sp->prev != NULL)
			fprintf(output_file, " -> ");
		fprintf(output_file, "%s", sp->s);
	}
	fprintf(output_file, "]\n");
}

static void
output_container_variables(char *name, struct variable_s *p)
{
	for (; p != NULL; p = p->next) {
		if (!strcmp(p->name, "short_description"))
			fprintf(output_file, "%s: %s\n", name, p->value);
		else if (!strcmp(p->name, "long_description"))
			fprintf(output_file, "%s\n", p->value);
		else
			fprintf(output_file, "%s: %s\n", p->name, p->value);
	}
}

static void
output_symbol_variables(char *name, struct variable_s *p)
{
	char *flags = NULL, *prototype = NULL;
	char *line = NULL, *remarks = NULL, *example = NULL;
	int noheader = 0;

	for (; p != NULL; p = p->next) {
		if (!strcmp(p->name, "header") && !strcmp(p->value, "no"))
			noheader = 1;
		else if (!strcmp(p->name, "d"))
			prototype = p->value;
		else if (!strcmp(p->name, "f"))
			flags = p->value;
		else if (!strcmp(p->name, "l"))
			line = p->value;
		else if (!strcmp(p->name, "r"))
			remarks = p->value;
		else if (!strcmp(p->name, "e"))
			example = p->value;
		else if (!strcmp(p->name, "t"))
			fprintf(output_file, "%s\n", p->value);
		else
			fprintf(output_file, "%s: %s\n", p->name, p->value);
	}

	if (flags != NULL) {
		int comma = 0;
		fprintf(output_file, "\nFlags: %s (", flags);
		while (*flags) {
			char *flag;
			if (comma)
				fprintf(output_file, ", ");
			switch (*flags++) {
			case 'b': flag = "builtin"; noheader = 1; break;
			case 'f': flag = "function"; break;
			case 'l': flag = "lvalue"; break;
			case 'm': flag = "macro"; break;
			case 's': flag = "struct"; break;
			case 't': flag = "type"; break;
			case 'u': flag = "union"; break;
			default: flag = "?";
			}
			comma = 1;
			fprintf(output_file, "%s", flag);
		}
		fprintf(output_file, ")\n");
	}

	if (line != NULL)
		fprintf(output_file, "\n%s\n", line);

	if (prototype != NULL) {
		fprintf(output_file, "\nDeclaration:\n");
		if (!noheader)
			fprintf(output_file, "#include <%s>\n\n", string_stack_tail->prev->s);
		fprintf(output_file, "%s\n", prototype);
	}

	if (remarks != NULL) {
		fprintf(output_file, "\nRemarks:\n");
		fprintf(output_file, "%s\n", remarks);
	}

	if (example != NULL) {
		fprintf(output_file, "\nExample:\n");
		fprintf(output_file, "%s\n", example);
	}
}

static int
_user_cmp(char *s1, char *s2)
{
	if (opt_pattern)
		return (pattern_match(s1, s2));
	else
		return (!strcmp(s1, s2));
}

static char *
strlower(char *s)
{
	char *p = s;
	for (; *p != '\0'; p++)
		*p = tolower(*p);
	return s;
}

static int
user_cmp(char *s1, char *s2)
{
	if (opt_case) {
		int retval;
		char *ps1, *ps2;
		if ((ps1 = (char *)malloc(strlen(s1) + 1)) == NULL)
			err(1, NULL);
		if ((ps2 = (char *)malloc(strlen(s2) + 1)) == NULL)
			err(1, NULL);
		strcpy(ps1, s1);
		strcpy(ps2, s2);
		strlower(ps1);
		strlower(ps2);
		retval = _user_cmp(ps1, ps2);
		free(ps1);
		free(ps2);
		return retval;
	} else
		return (_user_cmp(s1, s2));
}

static void
search_symbol(struct object_s *object_tree, char *symbol)
{
	struct object_s *p = object_tree;

	for (; p != NULL; p = p->next) {
		string_push(p->name);
		if (p->object_list != NULL) {
			if (user_cmp(p->name, symbol)) {
				output_string_stack_tree();
				if (!opt_list)
					output_container_variables(p->name, p->variable_list);
			}
			search_symbol(p->object_list, symbol);
		} else if (user_cmp(p->name, symbol)) {
			output_string_stack_tree();
			if (!opt_list)
				output_symbol_variables(p->name, p->variable_list);
		}
		string_pop();
	}
}

/*
 * Output the program syntax then exit.
 */
static void
usage(void)
{
	fprintf(stderr, "usage: cinfo [-cdlpsV] [-o file] [-L dir] [symbol...]\n");
	exit(1);
}

int
main(int argc, char **argv)
{
	int i, c;
	struct object_s *op;

	output_file = stdout;

	while ((c = getopt(argc, argv, "cdlo:psL:V")) != EOF)
		switch (c) {
		case 'c':
			opt_case = 1;
			break;
		case 'd':
			opt_debug = 1;
			break;
		case 'l':
			opt_list = 1;
			break;
		case 'o':
			if (output_file != stdout)
				fclose(output_file);
			if ((output_file = fopen(optarg, "w")) == NULL)
				err(1, "%s", optarg);
			break;
		case 'p':
			opt_pattern = 1;
			break;
		case 's':
			opt_pattern = 0;
			break;
		case 'L':
			opt_cinfodb_directory = 1;
			if ((opt_cinfodb_directory_arg = (char *)malloc(strlen(optarg) + 1)) == NULL)
				err(1, NULL);
			strcpy(opt_cinfodb_directory_arg, optarg);
			break;
		case 'V':
			fprintf(stderr, "%s - %s\n", version_string, rcsid);
			exit(0);
		case '?':
		default:
			usage();
			/* NOTREACHED */
		}
	argc -= optind;
	argv += optind;

	if (argc < 1)
		usage();

	if (!opt_cinfodb_directory) {
		char *p;
		if ((p = getenv("CINFODBDIR")) != NULL)
			op = make_object_tree(p);
		else
			op = make_object_tree(CINFODBDIR);
	} else
		op = make_object_tree(opt_cinfodb_directory_arg);

	for (i = 0; i < argc; i++)
		search_symbol(op, argv[i]);

	if (output_file != stdout)
		fclose(output_file);
	free_object_list(op);
	if (opt_cinfodb_directory)
		free(opt_cinfodb_directory_arg);

	return 0;
}
