/*
 * Copyright (c) 1995, 1996
 *	Sandro Sigala, Brescia, Italy.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Sandro Sigala.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: lexer.l,v 1.11 1996/06/29 13:21:40 sandro Exp $
 */

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "parser.h"

extern int opt_debug;
extern char *current_file_name;

#define MAX_INCLUDE_DEPTH 128

struct include_s {
	char *filename;
	char *path;
	int lineno;
	YY_BUFFER_STATE buffer_state;
};

struct include_s *include_stack[MAX_INCLUDE_DEPTH];

static int include_stack_idx = 0;

int lineno = 1;

static int yywrap(void);
static void string(void);
static void comment(void);
static void include(void);
static void include_push(char *);
static void include_pop(void);
%}

%%

\n			{ lineno++; }

"#"			{ comment(); }

".include"		{ include(); }

[a-zA-Z_][a-zA-Z_0-9]*	{ return IDENTIFIER; }

"\""			{ string(); return STRING; }

"("			{ return '('; }
")"			{ return ')'; }
"="			{ return '='; }

[ \t\v\f]+		/* eat up whitespaces */

<<EOF>>			{
				if (--include_stack_idx < 0)
					yyterminate();
				else
					include_pop();
			}

.			{ errx(1, "%s:%d: unrecognized character `%s'", current_file_name, lineno, yytext); }

%%

char *token_buffer;

/* max allowed token size */
static int maxtoken;

void
init_lex(void)
{
	maxtoken = 40;
	if ((token_buffer = (char *)malloc(maxtoken + 1)) == NULL)
		err(1, NULL);
}

void
done_lex(void)
{
	free(token_buffer);
}

static char *
extend_token_buffer(char *p)
{
	int offset = p - token_buffer;

	maxtoken = maxtoken * 2 + 10;

	if ((token_buffer = (char *)realloc(token_buffer, maxtoken + 2)) == NULL)
		err(1, NULL);

	return token_buffer + offset;
}

static int
yywrap(void)
{
	return 1;
}

static void
string(void)
{
	char *p;
	int c1;

	p = token_buffer;
	while ((c1 = input()) && c1 != '"' && c1 != EOF) {
		if (c1 == '\n')
			lineno++;
		if (p >= token_buffer + maxtoken)
			p = extend_token_buffer(p);
		if (c1 == '\\') {
			if ((c1 = input()) == '\n')
				lineno++;
			else
				*p++ = c1;
		} else
			*p++ = c1;
	}
	if (c1 == EOF)
		errx(1, "%s:%d: unexpected end of file in string", current_file_name, lineno);
	*p = '\0';
}

static void
comment(void)
{
	char c;
	while ((c = input()) != '\n' && c != EOF)
		;
	lineno++;
}


static char *
build_full_include_path(void)
{
	char *p;
	int i, strsize = 0;

	for (i = 0; i <= include_stack_idx; i++)
		if (include_stack[i]->path != NULL)
			strsize += strlen(include_stack[i]->path) + 1;

	if ((p = (char *)malloc(strsize + strlen(include_stack[include_stack_idx]->filename) + 1)) == NULL)
		err(1, NULL);

	*p = '\0';

	for (i = 0; i <= include_stack_idx; i++)
		if (include_stack[i]->path != NULL) {
			strcat(p, include_stack[i]->path);
			strcat(p, "/");
		}

	strcat(p, include_stack[include_stack_idx]->filename);

	return p;
}

static void
include(void)
{
	char *p;
	int c;

	while ((c = input()) == ' ' || c == '\t' || c == '\v' ||
	       c == '\f' || c == '\r')
	       	;

	if (c != '"')
		errx(1, "%s:%d: expected `\"' after include directive", current_file_name, lineno);

	p = token_buffer;
	while ((c = input()) != '"' && c != EOF) {
		if (p >= token_buffer + maxtoken)
			p = extend_token_buffer(p);
		*p++ = c;
	}
	if (c == EOF)
		errx(1, "%s:%d: unexpected end of file in string", current_file_name, lineno);
	*p = '\0';

	include_push(token_buffer);
}

static void
include_push(char *fname)
{
	char *p;

	if (include_stack_idx >= MAX_INCLUDE_DEPTH)
		errx(1, "%s:%d: includes nested too deeply", current_file_name, lineno);

	if ((include_stack[include_stack_idx] = (struct include_s *)malloc(sizeof(struct include_s))) == NULL)
		err(1, NULL);

	/*
	 * Check if the include directive changes the directory.
	 */
	if ((p = strchr(fname, '/')) != NULL) {
		if ((include_stack[include_stack_idx]->path = (char *)malloc(strlen(fname) - (p - fname) + 1)) == NULL)
			err(1, NULL);
		strncpy(include_stack[include_stack_idx]->path, fname, p - fname);
		include_stack[include_stack_idx]->path[p - fname] = '\0';
		if ((include_stack[include_stack_idx]->filename = (char *)malloc(strlen(++p) + 1)) == NULL)
			err(1, NULL);
		strcpy(include_stack[include_stack_idx]->filename, p);
	} else {
		include_stack[include_stack_idx]->path = NULL;
		if ((include_stack[include_stack_idx]->filename = (char *)malloc(strlen(fname) + 1)) == NULL)
			err(1, NULL);
		strcpy(include_stack[include_stack_idx]->filename, fname);
	}

	include_stack[include_stack_idx]->buffer_state = YY_CURRENT_BUFFER;
	include_stack[include_stack_idx]->lineno = lineno;

	p = build_full_include_path();

	free(current_file_name);
	if ((current_file_name = (char *)malloc(strlen(fname) + 1)) == NULL)
		err(1, NULL);
	strcpy(current_file_name, fname);

	if (opt_debug)
		printf("===> including %s\n", p);

	include_stack_idx++;

	if ((yyin = fopen(p, "r")) == NULL)
		err(1, "%s", p);
	free(p);

	lineno = 1;

	yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));

	BEGIN(INITIAL);
}

static void
include_pop(void)
{
	yy_delete_buffer(YY_CURRENT_BUFFER);

	yy_switch_to_buffer(include_stack[include_stack_idx]->buffer_state);
	free(current_file_name);

	if (opt_debug) {
	        char *p;
	        p = build_full_include_path();
		printf("<=== exiting %s\n", p);
		free(p);
        }

	if (include_stack[include_stack_idx]->path != NULL) {
		if ((current_file_name = (char *)malloc(strlen(include_stack[include_stack_idx]->path) + strlen(include_stack[include_stack_idx]->filename) + 1)) == NULL)
			err(1, NULL);
		strcpy(current_file_name, include_stack[include_stack_idx]->path);
	} else {
		if ((current_file_name = (char *)malloc(strlen(include_stack[include_stack_idx]->filename) + 1)) == NULL)
			err(1, NULL);
		*current_file_name = '\0';
	}
	strcat(current_file_name, include_stack[include_stack_idx]->filename);

	if (include_stack[include_stack_idx]->path != NULL)
		free(include_stack[include_stack_idx]->path);
	free(include_stack[include_stack_idx]->filename);

	lineno = include_stack[include_stack_idx]->lineno;

	free(include_stack[include_stack_idx]);
}
