/*
 * Copyright (c) 1995, 1996
 *	Sandro Sigala, Brescia, Italy.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Sandro Sigala.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: yyextract.c,v 1.8 1996/06/28 15:11:44 sandro Exp $
 */

static char *rcsid = "$Id: yyextract.c,v 1.8 1996/06/28 15:11:44 sandro Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <getopt.h>

#include "version.h"
#include "parser.h"
#include "tree.h"

extern struct object_s *parsing_tree;

extern int yyparse(void);
extern char *yytext;
extern FILE *yyin;

int parser_run;
int parser_eof;

enum {
	OPT_FORMAT_YACC = 0,
	OPT_FORMAT_BNF,
	OPT_FORMAT_EBNF
};

static int opt_format;

FILE *output_file;

static void
skip_section(int section)
{
	int tk;

	parser_run = 0;

	while ((tk = yylex()) != 0 && tk != SECTIONSEP)
		;

	if (section == 1) {
		if (tk == 0)
			errx(1, "unexpected end of file in the declaration section");
	} else if (tk == SECTIONSEP)
		errx(1, "found section separator in the additional c code section");
}

static void
output_components(struct object_s *p)
{
	struct object_s *last = NULL;
	int num = 0;

	fprintf(output_file, "\t\t");
	for (; p != NULL; p = p->next) {
		if (p->type == '|') {
			if (num == 0)
				fprintf(output_file, "/* empty */");
			fprintf(output_file, "\n\t\t| ");
			num = 0;
		} else {
			fprintf(output_file, "%s ", p->value);
			num++;
		}
		last = p;
	}
	if (last != NULL && last->type == '|' && num == 0)
		fprintf(output_file, "/* empty */");
}

static void
output_tree(struct object_s *p)
{
	int num = 0;
	for (; p != NULL; p = p->next) {
		if (num > 0)
			fprintf(output_file, "\n");
		fprintf(output_file, "%s:\n", p->value);
		if (p->assoc != NULL) {
			output_components(p->assoc);
			fprintf(output_file, "\n");
		}
		fprintf(output_file, "\t\t;\n");
		num++;
	}
}

static void
pspaces(int n)
{
	int i;
	for (i = 0; i < n; i++)
		fprintf(output_file, " ");
}


static void
bnf_output_components(struct object_s *p, int spaces)
{
	struct object_s *last = NULL;
	int num = 0;

	for (; p != NULL; p = p->next) {
		if (p->type == '|') {
			if (num == 0)
				fprintf(output_file, "/* empty */");
			fprintf(output_file, "\n");
			pspaces(spaces);
			fprintf(output_file, "| ");
			num = 0;
		} else {
			if (p->type == CHARACTER || p->type == STRING)
				fprintf(output_file, "%s ", p->value);
			else
				fprintf(output_file, "<%s> ", p->value);
			num++;
		}
		last = p;
	}
	if (last != NULL && last->type == '|' && num == 0)
		fprintf(output_file, "/* empty */");
}

static void
bnf_output_tree(struct object_s *p)
{
	int num = 0, i;
	for (; p != NULL; p = p->next) {
		if (num > 0)
			fprintf(output_file, "\n");
		i = fprintf(output_file, "<%s> ::= ", p->value);
		if (p->assoc != NULL)
			bnf_output_components(p->assoc, i);
		fprintf(output_file, "\n");
		num++;
	}
}

static int
bnf_find_empty(struct object_s *p)
{
	int num = 0;
	for (; p != NULL; p = p->next)
		if (p->type == '|') {
			if (num == 0)
				return 1;
			num = 0;
		} else
			num++;
	return (num == 0);
}

static void
ebnf_output_components(struct object_s *p, int spaces)
{
	int empty, num = 0;

	if ((empty = bnf_find_empty(p)))
		fprintf(output_file, "[ ");

	for (; p != NULL; p = p->next) {
		if (p->type == '|') {
			if (num > 0 && p->next != NULL) {
				fprintf(output_file, "\n");
				pspaces(spaces);
				fprintf(output_file, "| ");
			}
			num = 0;
		} else {
			if (p->type == CHARACTER || p->type == STRING)
				fprintf(output_file, "%s ", p->value);
			else
				fprintf(output_file, "<%s> ", p->value);
			num++;
		}
	}
	if (empty)
		fprintf(output_file, "]");
}


static void
ebnf_output_tree(struct object_s *p)
{
	int num = 0, i;
	for (; p != NULL; p = p->next) {
		if (num > 0)
			fprintf(output_file, "\n");
		i = fprintf(output_file, "<%s> ::= ", p->value);
		if (p->assoc != NULL)
			ebnf_output_components(p->assoc, i);
		fprintf(output_file, "\n");
		num++;
	}
}

static void
process_file(char *filename)
{
	if (filename != NULL && strcmp(filename, "-") != 0) {
		if ((yyin = fopen(filename, "r")) == NULL)
			err(1, "%s", filename);
	} else
		yyin = stdin;

	skip_section(1);

	parser_run = 1;
	parser_eof = 0;
	yyparse();
	if (!parser_eof)
		errx(1, "unexpected end of file in grammar section");

	skip_section(2);

	if (yyin != stdin)
		fclose(yyin);

	if (parsing_tree != NULL) {
		if (opt_format == OPT_FORMAT_YACC)
			output_tree(parsing_tree);
		else if (opt_format == OPT_FORMAT_BNF)
			bnf_output_tree(parsing_tree);
		else
			ebnf_output_tree(parsing_tree);
		free_object_list(parsing_tree);
	} else
		errx(1, "no grammar found");
}

static void
usage(void)
{
	fprintf(stderr, "usage: yyextract [-beyV] [-o file] [file...]\n");
	exit(1);
}

int
main (int argc, char **argv)
{
	int i, c;

	output_file = stdout;

	while ((c = getopt(argc, argv, "beo:yV")) != EOF)
		switch (c) {
		case 'b':
			opt_format = OPT_FORMAT_BNF;
			break;
		case 'e':
			opt_format = OPT_FORMAT_EBNF;
			break;
		case 'o':
			if (output_file != stdout)
				fclose(output_file);
			if ((output_file = fopen(optarg, "w")) == NULL)
				err(1, "%s", optarg);
			break;
		case 'y':
			opt_format = OPT_FORMAT_YACC;
			break;
		case 'V':
			fprintf(stderr, "%s - %s\n", version_string, rcsid);
			exit(0);
		case '?':
		default:
			usage();
			/* NOTREACHED */
		}
	argc -= optind;
	argv += optind;

	if (argc < 1)
		process_file(NULL);
	else
		for (i = 0; i < argc; i++)
			process_file(argv[i]);

	return 0;
}
